// $Header: /u/s/o/solomon/java/rooms.tmp/RCS/CalGrid.java,v 1.5 1996/06/20 19:23:44 solomon Exp solomon $
// Copyright (c) 1996, Marvin Solomon
package rooms;

import java.applet.Applet;
import java.awt.Canvas;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Event;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.util.Date;

// A CalGrid is a way of choosing a date from a month.
// It displays a 6 by 7 grid (see comment below) of numbers, with SMTWTFS
// column headers.

// At any one time, exactly one of the cells is "selected".  If one
// of the column headers is selected, this CalGrid is selecting a particular
// day of the week; otherwise, it is selecting a particular date.

// When a cell is clicked, that day or date is highlighted, and
// Rooms.changeDate() is called to inform it of the change in
// selection.
public class CalGrid extends Applet implements Constants {

	// Info about the current month
	int curYear;	// years since 1900
	int curMon;		// 0..11
	int maxDay;		// last day of month (e.g. Jan->31)
	int offset; 	// day of week of 1st day of month (e.g. Tues->2)

	// Info about the curent selection
	boolean dflag;		// if true, a day of the week rather than a date
	int curDay; 		// 0..6 if dflag==true, 1..31 if dflag==false

	// Dimensions
	// There are six rows because in the worst case, a month has four full
	// weeks and two partial weeks.  We could use fewer rows in some cases--as
	// few as four for a February starting on Sunday--but it's better to allow
	// enough room to change months without changing the size of this applet.
	int rows = 7;
	int cols = 7;

	// Various sizes, in pixels (there has to be a better way to do this!)
	int cheight, cwidth;	// cell dimensions
	int height, width;		// overall size

	static void debug(String s) { Rooms.debug(s); }

	public void init() {
		// Misc graphics setup
		Rooms.barrier.start(1);
		setBackground(Color.white);
		setForeground(Color.black);
		setFont(Rooms.defaultFont);

		// Calculate sizes
		FontMetrics fm = getFontMetrics(Rooms.defaultFont);
		cheight = fm.getHeight() + 2;
		cwidth = fm.stringWidth(" 00") + 2;
		width = cols*cwidth+10;
		height = rows*cheight + 2;

		// Initialize to today's date
		Date dt = new Date();

		curYear = dt.getYear();
		curMon = dt.getMonth();
		curDay = dt.getDate();
		maxDay = numDays[curMon];
		// Adjust for leap year (this will fail in year 2100!)
		if (curMon == 1 && (curYear % 4)==0) maxDay++;
		// offset = weekday(curYear,curMon,1);
		offset = new Date(curYear, curMon, 1).getDay();
		dflag = false;

		Rooms.calendar = this;
		Rooms.barrier.finish(1);
		repaint();
	}

	// Reset to the given month and year.  If we currently have selected a
	// day of the week, keep the same day of the week.  Otherwise choose
	// the first day or last day of the month according to whether the
	// change is forward or backward.
	public void setMonth(int m, int y) {
		maxDay = numDays[m];
		if (m == 1 && (y % 4)==0) maxDay++;
		if (!dflag) {
			if (y > curYear || (y == curYear && m > curMon))
				curDay = 1;
			else curDay = maxDay;
		}
		curYear = y;
		curMon = m;
		maxDay = numDays[curMon];
		if (curMon == 1 && (curYear % 4)==0) maxDay++;
		offset = new Date(curYear, curMon, 1).getDay();
		//debug("CalGrid.setMonth("+m+","+y+"), curYear="+y+", curMon="+m);
		repaint();
	}
	public String shortDate() {
		return dflag
			? dayname[curDay]
			: curMon+1 + "/" + curDay + "/" + curYear;
	}
	public String fullDate() {
		return dflag
			? dayname[curDay] + "s"
			: MonthChooser.monthName[curMon]
				+ " " + curDay + ", 19" + curYear;
	}

	// The following code computes the day-of week (Sunday==0) for day d
	// of month m in year 1900+y.  This code only works for m in the
	// range 1..12 and y in the range 70..199.
	// It is only needed because class Date is broken in the current
	// release of Netscape Navigator (2.0 SPARC Solaris).
	private static final int adj[] = { 0, 3, 3, 6, 1, 4, 6, 2, 5, 0, 3, 5 };
	private int weekday(int y, int m, int d) {
		// return new Date(y, m, d).getDay();  // This doesn't work.
		return 
			(4					// Jan 1, 1970 was a Friday.
				+ (y-70)		// add one day for each year
				+ (y-69)/4		// add one day for each preceding leap year
				+ adj[m]		// plus appropriate adjustment for month
				+ (((y%4) == 0 && m > 1) ? 1 : 0)	// one for Feb 29
				+ d-1
			) % 7;
	}

	// Helper function for paint() and mouseDown(), etc.
	// Translate a row/column hit into a day of the month (1..31), or
	// -1 if not a valid day.
	private int dayOf(int r, int c) {
		int n = cols*r + c + 1 - offset;
		if (n <= 0 || n > maxDay)
			return -1;
		return n;
	}

	// Called by the applet machinery to request a redraw.
	public void paint(Graphics g) {
		int x,y;

		// Set up colors and draw border
		g.setColor(Color.white);
		g.clearRect(0,0,size().width-1,size().height-1);
		g.setColor(Color.black);
		g.drawRect(0,0,size().width-1,size().height-1);

		// Fill in column headers
		for (int j=0; j<cols; j++) {
			x = j*cwidth+7;
			g.drawString(" S M T W T F S".substring(2*j,2*j+2),
				x, cheight-1);
			// indicate selected day
			if (dflag && curDay == j)
				g.drawRect(x-2, 3, cwidth+1, cheight-1);
		}
		// Fill in days of the month
		for (int i=1; i<rows; i++) {
			for (int j=0; j<cols; j++) {
				int n = dayOf(i-1,j);
				if (n < 0)
					continue;
				x = j*cwidth+7;
				y = i*cheight;
				g.drawString(
					(n>9 ? "" : " ") + n,
					x, y + cheight - 1);
				// indicate if selected date
				if (!dflag && n == curDay)
					g.drawRect(x-3, y+3, cwidth+1, cheight-1);
			}
		}
	}

	public Dimension minimumSize() { return new Dimension(width, height); }
	public Dimension preferredSize() { return minimumSize(); }

	// Respond to a mouse click.
	public boolean mouseDown(Event evt, int x, int y) {
		if (Rooms.disabled) return false;
		if (y < cheight) {
			// Click was in a column header
			dflag = true;
			curDay = x/cwidth;
		}
		else {
			y -= cheight;
			dflag = false;
			int n = dayOf(y/cheight, x/cwidth);
			if (n<0) {
				// Not a valid date.  Just ignore it.
				return false;
			}
			curDay = n;
		}
		Rooms.update();
		return false;
	}
} // CalGrid
