// $Header: /u/s/o/solomon/java/rooms.tmp/RCS/ReservationGrid.java,v 1.7 1996/06/20 19:23:44 solomon Exp solomon $
// Copyright (c) 1996, Marvin Solomon
package rooms;

import java.applet.Applet;
import java.awt.Frame;
import java.awt.Component;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Event;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.TextArea;
import java.util.Date;
import java.net.URL;
import java.net.URLConnection;
import java.io.InputStream;
import java.io.DataInputStream;
import java.io.IOException;

// A ReservationGrid shows all the active reservations for a given
// date.  Eventually, it will also allow you to add/delete reservations.
// The grid has columns corresponding to rooms and rows corresponding to
// time slots (see TIME_INTERVAL below for the precision).  There is
// also a header row and a header column.  Each reservation is
// indicated by a filled rectangle with a thick border.  There is optionally
// a "current" selected reservation, filled with a different color
// or a "proto-reservation" indicated by an unfilled thick border.
public class ReservationGrid extends Applet implements Constants {

	// A linked list of parsed reservations returned by the server.
	Reservation resList = null;
	// Currently selected reservation (if any).
	Reservation current = null;
	// Current proto-reservation (if any -- see below).
	int PRcol = -1, PRmin, PRmax;

	// Off-screen image to hold relatively static portion of window
	private Image grid;
	private Graphics gridGraph; // grid.getGraphics()

	void debug(String s) { Rooms.debug(s); }

	public void init() {
		Rooms.barrier.start(3);
		setBackground(Color.white);
		setForeground(Color.black);
		Rooms.grid = this;

		// start things up
		Rooms.disabled = false;
		Rooms.barrier.finish(3);
		Rooms.update();
	}

	// Reload list of reservations for the indicated date.
	public void setDate(String d) {
		String cmdLine = Rooms.cgiBin + "/rooms.pl?"
			/* Add this for debugging the server
			+ (Rooms.debugFlag
				? "-D+/u/s/o/solomon/junk/log+-h+gjetost+"
			 	: "")
			*/
			+ "print+-v+-J+" + d;
			// The server cgi script should be invoked as
			//   "rooms -v -J 3/11/96"
			// or
			//   "rooms -v -J monday"
			// The args are encoded as part of the URL using the cgi
			// convention:  Args are preceded by '?' and separated by '+'.

		// Create a URL connects to the server and send the arguments.
		URL url;
		try {
			debug("connecting to "+cmdLine);
			url = new URL(cmdLine);
		}
		catch (Throwable e) {
			debug("exception connecting to "+cmdLine+": "+e);
			e.printStackTrace();
			return;
		}

		// Build a list of reservations returned.
		// The "-J" causes the server to encode the response in a way
		// that is particularly easy for us to parse.
		resList = null;
		try {
			// Create a stream to read the response
			URLConnection conn = url.openConnection();
			conn.setUseCaches(false);
			DataInputStream str = new DataInputStream(conn.getInputStream());
			// Create a Parser to parse
			// the response, sending the results back to us.
			Parser.parse(str, this);
		}
		catch (Throwable e) {
			System.out.println("can't open url: "+e);
			e.printStackTrace();
			return;
		}

		// Initially, no reservation is selected, so fill the TextArea
		// with an informational message.
		current = null;
		PRcol = -1;
		Rooms.clearChoice();
		Rooms.clearSelection();
		Rooms.info.setText(DEFAULT_MSG);
		grid = null;
		repaint();
	} // ReservationGrid.setDate

	public Dimension minimumSize() {
		return new Dimension(GRID_WIDTH, GRID_HEIGHT);
	}
	public Dimension preferredSize() { return minimumSize(); }

	// Helper functions.

	// Arg t is a time expressed as a number of slots after START_HR.
	// Result is 12-hr time in the form 09:45
	private String slot2str(int t) {
		t = 60*START_HR + t*TIME_INTERVAL;	// convert to minutes past midnight
		if (t >= 60*13) {
			// t -= 60*12;	// convert to 12-hour time
			// NB: bug in Java interpreter in Netscape 3.04b
			t = t - 60*12;	// convert to 12-hour time
		}
		int h = t/60;
		t %= 60;
		return (h < 10 ? " "+h : ""+h)
			+ ":"
			+ (t < 10 ? "0"+t : ""+t);
	} // ReservationGrid.slot2str

	// Translate from room name to column number
	private int lookupRoom(String s) {
		int i;
		for (i=0; i<roomName.length; i++)
			if (roomName[i].equals(s)) return i;
		return -1;
	} // ReservationGrid.lookupRoom

	// Arg s is a character string containing a number of seconds past
	// midnight (in decimal).  Result is a slot number.
	private int str2time(String s) throws NumberFormatException {
		int n = Integer.parseInt(s);
		n /= 60*TIME_INTERVAL;				// convert to ticks
		n -= START_HR*(60/TIME_INTERVAL);	// shift to start of grid
		if (n < 0)
			n = 0;					// clip to range
		if (n >= TIME_PERIODS)
			n = TIME_PERIODS-1;		// clip to range
		return n;
	} // ReservationGrid.str2time

	// Mark off a block of time in the grid as a rectangle with thicker
	// borders and a fill color of `color'.  If the `color' argument is
	// null, don't fill, but make the borders even thicker.
	private void drawBlock(Graphics g,
			int col, int minRow, int maxRow, Color color)
	{
		int x = TIME_PERIOD_WIDTH + ROOM_WIDTH * col;
		int y = ROOM_NAME_HEIGHT + TIME_SLOT_HEIGHT * minRow;
		int w = ROOM_WIDTH;
		int h = TIME_SLOT_HEIGHT * (maxRow - minRow);
		if (color != null) {
			g.setColor(color);
			g.fillRect(x, y, w, h);
		}
		g.setColor(Color.black);
		g.drawRect(x, y, w, h);
		g.drawRect(x+1, y+1, w-2, h-2);
		if (color == null)
			g.drawRect(x+2, y+2, w-4, h-4);
	}

	// Actually display stuff to the window
	public void update(Graphics g) {
		int x,y,w,h;
		Dimension d = size();

		if (grid == null) {
			grid = createImage(d.width, d.height);
			gridGraph = grid.getGraphics();
			gridGraph.setFont(Rooms.smallFont);
			showReservations(gridGraph);
		}
		g.drawImage(grid, 0, 0, this);

		// Indicate selected reservation (if any).
		if (current != null)
			drawBlock(g,
				current.room, current.start, current.end, SEL_RES_COLOR);

		// Show the current proto-reservation (if any)
		if (PRcol != -1)
			drawBlock(g, PRcol, PRmin, PRmax+1, null);
	}
		
	public void paint(Graphics g) {
		update(g);
	}

	// Display the relatively "static" portions of the grid.
	// When this is called, g is an offscreen imaage.
	private void showReservations(Graphics g) {
		int x, y, h, w, i, swidth;
		FontMetrics fm = g.getFontMetrics();

		// Outer box
		g.setColor(Color.black);
		g.drawRect(TIME_PERIOD_WIDTH, 0,
			GRID_WIDTH - TIME_PERIOD_WIDTH - 1, GRID_HEIGHT - 1);

		// Horizontal lines
		for (i=0; i < TIME_PERIODS; i++) {
			y = ROOM_NAME_HEIGHT + TIME_SLOT_HEIGHT*i;
			if (((i * TIME_INTERVAL) % 30) != 0)
				g.setColor(Color.lightGray);
			g.drawLine( TIME_PERIOD_WIDTH, y, GRID_WIDTH-1, y);
			g.setColor(Color.black);
			if (((i * TIME_INTERVAL) % 60) == 0)
				g.drawLine( TIME_PERIOD_WIDTH, y+1, GRID_WIDTH-1, y+1);
		}

		// Vertical lines
		for (i=0; i < roomName.length; i++) {
			g.drawLine(
				TIME_PERIOD_WIDTH + ROOM_WIDTH*i, 0,
				TIME_PERIOD_WIDTH + ROOM_WIDTH*i, GRID_HEIGHT);
		}

		// Room names
		y = ROOM_NAME_HEIGHT/2 + 3;
		for (i = 1; i <= roomName.length; i++) {
			swidth = fm.stringWidth(roomName[i-1]);
			x = TIME_PERIOD_WIDTH + ROOM_WIDTH*(i-1) + (ROOM_WIDTH - swidth)/2;
			g.drawString(roomName[i-1], x, y);
		}

		// Time periods
		swidth = fm.stringWidth("00:00");
		x = (TIME_PERIOD_WIDTH - swidth)/2;
		if (x < 0) x = 0;
		y = ROOM_NAME_HEIGHT + TIME_SLOT_HEIGHT/2 + 2;
		for (i=0; i < TIME_PERIODS; i+=2) {
			g.drawString(slot2str(i), x, y);
			y += 2*TIME_SLOT_HEIGHT;
		}

		// Reservations
		for (Reservation r = resList; r!=null; r = r.next)
			drawBlock(g, r.room, r.start, r.end, RES_COLOR);
	} // ReservationGrid.showReservations

	// Mouse event handling for selecting existing reservations or
	// creating new ones.  Any any time, there is optionally a
	// "proto-reservation" (PR), which consists of a contiguous set of cells
	// in one column not overlapping with any current reservation.
	// The PR is recorded in variables PRcol, PRmin, and PRmax,
	// representing its column, first row, and last row.  The special value
	// PRcol == -1 indicates that there is currently no PR.
	// The current PR is turned off by any click, or by a drag
	// that changes columns, moves into an existing reservation, or moves
	// above the top of the current PR.
	// (The same drag may create a new PR as described below).
	// A click or drag into an existing reservation selects it.  A click
	// in an unoccupied cell, or a drag into an unoccupied cell when no
	// PR is active creates a new one-cell proto-res.  A drag
	// past the end of the current proto-res (in the same column) extends it.
	// These rather complicated rules seem to give the right feel.
	public boolean handleEvent(Event evt) {
		if (Rooms.disabled) return false;
		// Translate from pixels to row/column
		int x = (evt.x - TIME_PERIOD_WIDTH)/ROOM_WIDTH;
		int y = (evt.y - ROOM_NAME_HEIGHT)/TIME_SLOT_HEIGHT;
		boolean changed = false;

		// We're only interested in down and drag events
		if (evt.id != Event.MOUSE_DOWN && evt.id != Event.MOUSE_DRAG)
			return false;

		// Make sure we're inside the body of the grid
		if (x < 0 || x >= roomName.length || y < 0 || y >= TIME_PERIODS) {
			return false;
		}

		// Find a reservation that covers the selected box if any.
		Reservation r;
		for (r = resList; r!=null; r=r.next) {
			if (r.hits(y,x))
				break;
		}

		// Now implement the above rules
		if (r != null) {
			// Existing reservation.

			if (r != current) {
				// Copy the reservation's comment into the text window.
				Rooms.info.setText("");
				Rooms.info.appendText(r.detail);
			}
			Rooms.clearChoice();
			Rooms.setSelection(r);
			changed =  (current != r || PRcol != -1);
			current = r;
			PRcol = -1;
		}
		else if (Rooms.ticket == null)
			return false;  // not allowed to make reservations
		else {
			// Empty cell

			if (current != null) {
				Rooms.info.setText(DEFAULT_MSG);
				current = null;
			}
			int oldCol = PRcol, oldMin = PRmin, oldMax = PRmax;
			// If we are not extending the previous value, clear it away
			if (PRcol != -1 &&
				(evt.id == Event.MOUSE_DOWN || x != PRcol || y < PRmin))
			{
				PRcol = -1;
			}
			// Then set or extend it to the new cell
			if (PRcol == -1) {
				// a new PR
				PRcol = x;
				PRmin = PRmax = y;
			}
			else if (y >= PRmin)
				PRmax = y;
			changed = (current != null
				|| oldCol != PRcol
				|| oldMin != PRmin
				|| oldMax != PRmax);
			if (changed)
				Rooms.setChoice(roomName[PRcol],
						printTime(PRmin),
						printTime(PRmax+1));
			Rooms.clearSelection();
		}
		if (changed)
			repaint();
		return false;
	} // handleEvent

	// Used by Parser to add another reservation to this ReservationGrid
	void addReservation(String id, String sroom,
		String start, String end, String detail)
	{
		int room = lookupRoom(sroom);
		if (room < 0)
			return;
		resList = new Reservation(
			id, room, str2time(start), str2time(end), detail, resList);
	} // ReservationGrid.addReservation

	String printTime(int slot) {
		int minutes = slot * TIME_INTERVAL;
		int hours = minutes/60 + START_HR;
		minutes %= 60;
		return "" + hours + ":" + (minutes < 10 ? "0" : "") + minutes;
	}
} // ReservationGrid
